@@ -1,6 +1,6 @@
 {-# LANGUAGE Arrows #-}
 
-{- 
+{-
 ParseSyntaxFiles.hs processes a directory containing Kate
 XML syntax highlighting definitions.  For each xml file in
 the directory, it creates a syntax highlighting parser
@@ -21,23 +21,20 @@ Requires HXT.
 module Main where
 
 import Text.XML.HXT.Core
-import Control.Arrow
-import Control.Arrow.ArrowList
-import Control.Monad (liftM)
+import Control.Monad
 import Data.List
-import Data.Maybe
 import Data.Char (toUpper, toLower, isAlphaNum)
 import qualified Data.Map as Map
 import System.Directory
 import System.Environment
-import System.Exit
 import System.FilePath
 import Text.PrettyPrint
 import Text.Printf (printf)
 import Data.Char (ord)
-import Text.Highlighting.Kate.Definitions
+import Text.Highlighting.Kate.Types
 import qualified Data.ByteString as B
 import Data.ByteString.UTF8 (fromString, toString)
+import Text.Regex.Posix ((=~))
 
 data SyntaxDefinition =
   SyntaxDefinition { synLanguage      :: String
@@ -57,7 +54,7 @@ data SyntaxKeywordAttr =
                      , keywordDelims          :: [Char]
                      } deriving (Read, Show)
 
-data SyntaxContext = 
+data SyntaxContext =
   SyntaxContext { contName               :: String
                 , contAttribute          :: String
                 , contLineEndContext     :: String
@@ -65,7 +62,7 @@ data SyntaxContext =
                 , contFallthrough        :: Bool
                 , contFallthroughContext :: String
                 , contDynamic            :: Bool
-                , contParsers            :: [SyntaxParser] 
+                , contParsers            :: [SyntaxParser]
                 } deriving (Read, Show)
 
 data SyntaxParser =
@@ -83,156 +80,182 @@ data SyntaxParser =
                , parserChildren          :: [SyntaxParser]
                } deriving (Read, Show)
 
--- | Converts a list of files (ending in .xml) and directories containing .xml files
--- into a list of .xml files.
+-- | Converts a list of files (ending in .xml) and directories containing
+-- .xml files into a list of .xml files.
 argFiles :: [String] -> IO [String]
-argFiles [] = error "Specify paths of xml files and/or directories containing xml syntax files."
+argFiles [] = error "Specify paths of xml files and/or directories."
 argFiles args = do
   let isXmlFile x = isSuffixOf ".xml" x
   let (files, dirs) = partition isXmlFile args
-  dirContents <- mapM (\dir -> getDirectoryContents dir >>= return . map (combine dir) . filter isXmlFile) dirs
+  dirContents <- forM dirs $ \dir -> do
+                   dc <- getDirectoryContents dir
+                   return $ map (combine dir) $ filter isXmlFile dc
   return $ nub (files ++ concat dirContents)
 
-
+libraryPath :: FilePath
 libraryPath = joinPath ["Text", "Highlighting", "Kate"]
+
+destDir :: FilePath
 destDir = joinPath [libraryPath, "Syntax"]
 
 main :: IO ()
 main = do
   files <- getArgs >>= argFiles
   destDirExists <- doesDirectoryExist destDir
-  if destDirExists
-     then return ()
-     else createDirectory destDir 
+  unless destDirExists $ createDirectory destDir
   mapM_ processOneFile files
+  names <- (sort . map dropExtension . filter (isSuffixOf ".hs"))
+          `fmap` (getDirectoryContents destDir)
+  writeSyntaxFile names
+  writeCabalFile names
+
+writeSyntaxFile :: [String] -> IO ()
+writeSyntaxFile names = do
   let syntaxFile = combine libraryPath (addExtension "Syntax" "hs")
   putStrLn $ "Writing " ++ syntaxFile
   -- Get all syntax files, not only the newly generated ones.
-  names <- getDirectoryContents destDir >>= return . sort . map dropExtension . filter (isSuffixOf ".hs")
-  let imports = unlines $ map (\name -> "import qualified Text.Highlighting.Kate.Syntax." ++ name ++ " as " ++ name) names 
-  let cases = unlines $ map (\name -> show (map toLower name) ++ " -> " ++ name ++ ".highlight") names
-  let languageExtensions = '[' :
-        (intercalate ", " $ map (\name -> "(" ++ show name ++ ", " ++ name ++ ".syntaxExtensions)") names) ++ "]"
+  let imports = unlines $ map (\name ->
+                 "import qualified Text.Highlighting.Kate.Syntax." ++ name ++
+                 " as " ++ name) names
+  let cases = unlines $ map (\name ->
+                 show (map toLower name) ++ " -> " ++ name ++ ".highlight")
+                 names
+  let languageExtensions = "[" ++ (intercalate ", " $ map (\name ->
+        "(" ++ show name ++ ", " ++ name ++ ".syntaxExtensions)") names) ++ "]"
   syntaxFileTemplate <- liftM toString $ B.readFile (syntaxFile <.> "in")
-  let filledTemplate = fillTemplate 0 [("imports",imports),
-                                       ("languages",show names),
-                                       ("languageExtensions",languageExtensions),
-                                       ("cases",cases)] syntaxFileTemplate
+  let filledTemplate = fillTemplate 0
+                          [("imports",imports),
+                           ("languages",show names),
+                           ("supportedlanguages", intercalate ", " $ map (\x ->
+                                "@" ++ map toLower x ++ "@") names),
+                           ("languageExtensions",languageExtensions),
+                           ("cases",cases)] syntaxFileTemplate
   B.writeFile syntaxFile $ fromString filledTemplate
 
+writeCabalFile :: [String] -> IO ()
+writeCabalFile names = do
+  copyFile "highlighting-kate.cabal" "highlighting-kate.cabal.orig"
+  cabalLines <- lines `fmap` readFile "highlighting-kate.cabal.orig"
+  let (front,rest) = break (=~ "Text\\.Highlighting\\.Kate\\.Syntax\\.")
+                     cabalLines
+  let end = dropWhile (=~ "Text\\.Highlighting\\.Kate\\.Syntax\\.") rest
+  let toMod n = replicate 21 ' ' ++ "Text.Highlighting.Kate.Syntax." ++ n
+  let newCabalLines = front ++ (map toMod names) ++ end
+  writeFile "highlighting-kate.cabal" $ unlines newCabalLines
+  putStrLn "Modified highlighting-kate.cabal."
+  putStrLn "Backed up original as highlighting-kate.cabal.orig."
+
 processOneFile :: FilePath -> IO ()
 processOneFile src = do
   [syntax] <- runX $ application src
   let name = nameFromPath src
   let outFile = joinPath [libraryPath, "Syntax", addExtension name "hs"]
+  let isIncludeRules p = parserType p == "IncludeRules" && "##" `isPrefixOf`
+                              parserContext p
   let includeLangs = nub $ filter (/= name) $ map (drop 2 . parserContext) $
-        filter (\p -> (parserType p) == "IncludeRules" && "##" `isPrefixOf` (parserContext p)) $ 
-        concatMap contParsers $ synContexts syntax
-  let includeImports = map (("import qualified " ++) . langNameToModule) includeLangs
+        filter isIncludeRules $ concatMap contParsers $ synContexts syntax
+  let includeImports = map (("import qualified " ++) . langNameToModule)
+                        includeLangs
   putStrLn $ "Writing " ++ outFile
   B.writeFile outFile $ fromString $
-           "{- This module was generated from data in the Kate syntax highlighting file " ++ (takeFileName src) ++ ", version " ++ 
-           synVersion syntax ++ ",\n" ++
-           "   by  " ++ synAuthor syntax ++ " -}\n\n" ++ 
-           "module Text.Highlighting.Kate.Syntax." ++ name ++ " ( highlight, parseExpression, syntaxName, syntaxExtensions ) where\n\
-           \import Text.Highlighting.Kate.Definitions\n\
+           "{- This module was generated from data in the Kate syntax\n\
+           \   highlighting file " ++ (takeFileName src) ++ ", version " ++
+           synVersion syntax ++ ", by " ++ synAuthor syntax ++ " -}\n\n" ++
+           "module Text.Highlighting.Kate.Syntax." ++ name ++ "\n          " ++
+           "(highlight, parseExpression, syntaxName, syntaxExtensions)" ++
+           "\nwhere\n\
+           \import Text.Highlighting.Kate.Types\n\
            \import Text.Highlighting.Kate.Common\n" ++
-           unlines includeImports ++ 
-           "import Text.ParserCombinators.Parsec\n\
-           \import Control.Monad (when)\n\
+           unlines includeImports ++
+           "import Text.ParserCombinators.Parsec hiding (State)\n\
            \import Data.Map (fromList)\n\
-           \import Data.Maybe (fromMaybe, maybeToList)\n\n" ++
+           \import Control.Monad.State\n\
+           \import Data.Char (isSpace)\n\
+           \import Data.Maybe (fromMaybe)\n" ++
            (if null (synLists syntax)
-               then ""
-               else "import qualified Data.Set as Set\n") ++
+               then "\n"
+               else "import qualified Data.Set as Set\n\n") ++
            render (mkParser syntax) ++ "\n"
 
+labelFor :: SyntaxDefinition -> String -> TokenType
+labelFor syntax attr' =
+  case lookup attr' (synItemDatas syntax) of
+       Just "dsKeyword" -> KeywordTok
+       Just "dsDataType" -> DataTypeTok
+       Just "dsDecVal" -> DecValTok
+       Just "dsBaseN" -> BaseNTok
+       Just "dsFloat" -> FloatTok
+       Just "dsChar" -> CharTok
+       Just "dsString" -> StringTok
+       Just "dsComment" -> CommentTok
+       Just "dsOthers" -> OtherTok
+       Just "dsAlert" -> AlertTok
+       Just "dsFunction" -> FunctionTok
+       Just "dsRegionMarker" -> RegionMarkerTok
+       Just "dsError" -> ErrorTok
+       _ -> NormalTok
+
 mkParser :: SyntaxDefinition -> Doc
-mkParser syntax = 
+mkParser syntax =
   let name = text "-- | Full name of language." $$
              text "syntaxName :: String" $$
              text ("syntaxName = " ++ show (synLanguage syntax))
       exts = text "-- | Filename extensions for this language." $$
              text "syntaxExtensions :: String" $$
              text ("syntaxExtensions = " ++ show (synExtensions syntax))
-      shortFormOf "dsKeyword" = ["kw"]
-      shortFormOf "dsDataType" = ["dt"] 
-      shortFormOf "dsDecVal" = ["dv"]
-      shortFormOf "dsBaseN" = ["bn"]
-      shortFormOf "dsFloat" = ["fl"]
-      shortFormOf "dsChar" = ["ch"]
-      shortFormOf "dsString" = ["st"]
-      shortFormOf "dsComment" = ["co"]
-      shortFormOf "dsOthers" = ["ot"]
-      shortFormOf "dsAlert" = ["al"]
-      shortFormOf "dsFunction" = ["fu"]
-      shortFormOf "dsRegionMarker" = ["re"]
-      shortFormOf "dsError" = ["er"]
-      shortFormOf _ = []
-      styles = text ("styles = " ++ (show [(typ, shortsty) | (typ, sty) <- synItemDatas syntax, shortsty <- shortFormOf sty]))
       withAttr = text "withAttribute attr txt = do" $$ (nest 2 $
                    text "when (null txt) $ fail \"Parser matched no text\"" $$
-                   text "let labs = attr : maybeToList (lookup attr styles)" $$
-                   text "st <- getState" $$
-                   text "let oldCharsParsed = synStCharsParsedInLine st" $$
-                   text "let prevchar = if null txt then '\\n' else last txt" $$
-                   text "updateState $ \\st -> st { synStCharsParsedInLine = oldCharsParsed + length txt, synStPrevChar = prevchar } " $$
-                   text "return (labs, txt)")
-      parseExpressionInternal = text "parseExpressionInternal = do" $$ (nest 2 $ 
+                   text "updateState $ \\st -> st { synStPrevChar = last txt" $$
+                   text "                        , synStPrevNonspace = synStPrevNonspace st || not (all isSpace txt) }" $$
+                   text "return (attr, txt)")
+      parseExpressionInternal = text "parseExpressionInternal = do" $$ (nest 2 $
                                   text "context <- currentContext" $$
-                                  text "parseRules context <|> (pDefault >>= withAttribute (fromMaybe \"\" $ lookup context defaultAttributes))")
+                                  text "parseRules context <|> (pDefault >>= withAttribute (fromMaybe NormalTok $ lookup context defaultAttributes))")
       parseExpression = text "-- | Parse an expression using appropriate local context." $$
-                        text "parseExpression :: GenParser Char SyntaxState LabeledSource" $$
-                        text "parseExpression = do" $$ (nest 2 $ 
+                        text "parseExpression :: KateParser Token" $$
+                        text "parseExpression = do" $$ (nest 2 $
                           text "st <- getState" $$
                           text "let oldLang = synStLanguage st" $$
                           text ("setState $ st { synStLanguage = " ++ show (synLanguage syntax) ++ " }") $$
                           text ("context <- currentContext <|> (pushContext " ++ show (contName $ head $ synContexts syntax) ++
                                 " >> currentContext)") $$
                           text "result <- parseRules context" $$
+                          text "optional $ eof >> pEndLine" $$
                           text "updateState $ \\st -> st { synStLanguage = oldLang }" $$
                           text "return result")
-      defaultAttributes = text $ "defaultAttributes = " ++ (show $ map (\cont -> (contName cont, contAttribute cont)) $ synContexts syntax)
+      defaultAttributes = text $ "defaultAttributes = " ++ (show $ map (\cont -> (contName cont, labelFor syntax $ contAttribute cont)) $ synContexts syntax)
       -- Note: lineBeginContexts seems not to be used in any of the xml files
-      -- lineBeginContexts = 
+      -- lineBeginContexts =
       --   text $ "lineBeginContexts = " ++ (show $ map (\cont -> (contName cont, contLineBeginContext cont)) $ synContexts syntax)
       startingContext = head (synContexts syntax)
+      contextNull = text $ "parseRules \"\" = parseRules " ++ show (contName startingContext)
       contextCatchAll = text $ "parseRules x = fail $ \"Unknown context\" ++ x"
       contexts = map (mkRules syntax) $ synContexts syntax
       initialContextStack = Map.fromList [(synLanguage syntax, [contName startingContext])]
       startingState = SyntaxState { synStContexts = initialContextStack
                                   , synStLanguage = synLanguage syntax
-                                  , synStCurrentLine = ""
-                                  , synStCharsParsedInLine = 0
+                                  , synStLineNumber = 0
+                                  , synStPrevNonspace = False
                                   , synStPrevChar = '\n'
                                   , synStCaseSensitive = synCaseSensitive syntax
                                   , synStKeywordCaseSensitive = keywordCaseSensitive (synKeywordAttr syntax)
                                   , synStCaptures = [] }
       initState = text $ "startingState = " ++ show startingState
-      sourceLineParser = text "parseSourceLine = manyTill parseExpressionInternal pEndLine"
-      mainParser = text "parseSource = do " $$  
-                   (nest 2 $ text "lineContents <- lookAhead wholeLine" $$
-                             text "updateState $ \\st -> st { synStCurrentLine = lineContents }" $$
-                             -- text "context <- currentContext" $$
-                             -- text "pushContext (fromMaybe \"#stay\" $ lookup context lineBeginContexts)" $$
-                             text "result <- manyTill parseSourceLine eof" $$
-                             text "return $ map normalizeHighlighting result")
       mainFunction = text $ "-- | Highlight source code using this syntax definition.\n\
-                            \highlight :: String -> Either String [SourceLine]\n\
-                            \highlight input =\n\
-                            \  case runParser parseSource startingState \"source\" input of\n\
-                            \    Left err     -> Left $ show err\n\
-                            \    Right result -> Right result"
+                            \highlight :: String -> [SourceLine]\n\
+                            \highlight input = evalState (mapM parseSourceLine $ lines input) startingState"
+      lineParser = text   $ "parseSourceLine :: String -> State SyntaxState SourceLine\n\
+                            \parseSourceLine = mkParseSourceLine parseExpressionInternal pEndLine"
       endLineParser = text "pEndLine = do" $$
-                      (nest 2 $ text "lookAhead $ newline <|> (eof >> return '\\n')" $$
+                      (nest 2 $ text "updateState $ \\st -> st{ synStPrevNonspace = False }" $$
                                 text "context <- currentContext" $$
                                 text "case context of" $$
-                                (nest 2 $ (vcat $ map (\cont -> text (show $ contName cont) <> text " -> " <> 
-                                            switchContext (contLineEndContext cont) <>
+                                (nest 2 $ (vcat $ map (\cont -> text (show $ contName cont) <> text " -> " <>
+                                            switchContext (contLineEndContext cont) (<> text " >> ") <>
                                             if "#pop" `isPrefixOf` (contLineEndContext cont)
-                                               then text " >> pEndLine"
-                                               else text " >> pHandleEndLine") $ synContexts syntax) $$
-                                          (text $ "_ -> pHandleEndLine")))
+                                               then text "pEndLine"
+                                               else text "return ()") $ synContexts syntax) $$
+                                          (text $ "_ -> return ()")))
                                 {- text "pushContext (fromMaybe \"#stay\" $ lookup context lineBeginContexts)" $$ -}
       -- we use 'words "blah blah2 blah3"' to keep ghc from inlining the list, which makes compiling take a long time
       listDef (n, list) = text $ listName n ++ " = Set.fromList $ words $ " ++
@@ -243,12 +266,12 @@ mkParser syntax =
       regexDef re = text $ compiledRegexName re ++ " = compileRegex " ++ show re
       regexes = vcat $ map regexDef $ nub $ [parserString x | x <- concatMap contParsers (synContexts syntax),
                                                               parserType x == "RegExpr", parserDynamic x == False]
-  in  vcat $ intersperse (text "") $ [name, exts, mainFunction, parseExpression, mainParser, initState, sourceLineParser, 
-                                      endLineParser, withAttr, styles, parseExpressionInternal, lists, regexes,
-                                      defaultAttributes {- , lineBeginContexts -}] ++ contexts ++ [contextCatchAll]
+  in  vcat $ intersperse (text "") $ [name, exts, mainFunction, lineParser, parseExpression, initState,
+                                      endLineParser, withAttr, parseExpressionInternal, lists, regexes,
+                                      defaultAttributes {- , lineBeginContexts -}] ++ contexts ++ [contextNull, contextCatchAll]
 
 mkAlternatives :: [Doc] -> Doc
-mkAlternatives docs = 
+mkAlternatives docs =
   let contents = vcat $ intersperse (text "<|>") docs
   in  if length docs > 1
          then char '(' <> contents <> char ')'
@@ -257,24 +280,25 @@ mkAlternatives docs =
 mkRules :: SyntaxDefinition -> SyntaxContext -> Doc
 mkRules syntax context =
   let fallthroughParser = if contFallthrough context
-                             then [parens (switchContext (contFallthroughContext context) <> 
-                                   text " >> return ([], \"\")")]
+                             then [parens (switchContext (contFallthroughContext context) (<> text " >> ") <>
+                                   text "currentContext >>= parseRules")]
                              else []
-  in  text ("parseRules " ++ show (contName context) ++ " = ") $$ 
+  in  text ("parseRules " ++ show (contName context) ++ " =") $$
       if null (contParsers context) && null fallthroughParser
          then nest 2 (text "pzero")
-         else nest 2 $ (text "do (attr, result) <- " <> 
-                       (mkAlternatives $ (map (mkSyntaxParser syntax context) $ contParsers context) ++ fallthroughParser)) $$
-                       text ("   return (attr, result)")
+         else nest 2 $ mkAlternatives $ (map (mkSyntaxParser syntax context) $ contParsers context) ++ fallthroughParser
 
 mkSyntaxParser :: SyntaxDefinition -> SyntaxContext -> SyntaxParser -> Doc
-mkSyntaxParser syntax context parser = 
-  let mainParser = text $ case parserType parser of
+mkSyntaxParser syntax context parser =
+  let attr' = case parserAttribute parser of
+                   "" -> labelFor syntax $ contAttribute context
+                   x  -> labelFor syntax x
+      mainParser = text $ case parserType parser of
             "DetectChar"       -> "pDetectChar " ++ show (parserDynamic parser) ++ " " ++ show (parserChar parser)
-            "Detect2Chars"     -> "pDetect2Chars " ++ show (parserDynamic parser) ++ " " ++ 
+            "Detect2Chars"     -> "pDetect2Chars " ++ show (parserDynamic parser) ++ " " ++
                                     show (parserChar parser) ++ " " ++ show (parserChar1 parser)
             "AnyChar"          -> "pAnyChar " ++ show (parserString parser)
-            "StringDetect"     -> "pString " ++ show (parserDynamic parser) ++ " " ++ show (parserString parser) 
+            "StringDetect"     -> "pString " ++ show (parserDynamic parser) ++ " " ++ show (parserString parser)
             "RegExpr"          -> if parserDynamic parser
                                      then "pRegExprDynamic " ++ show (parserString parser)
                                      else "pRegExpr " ++ compiledRegexName (parserString parser)
@@ -293,9 +317,9 @@ mkSyntaxParser syntax context parser =
             "LineContinue"     -> "pLineContinue"
             "IncludeRules"     -> case parserContext parser of
                                       ('#':'#':xs) -> langNameToModule xs ++ ".parseExpression" ++
-                                                      if parserIncludeAttrib parser || null (parserAttribute parser)
+                                                      if parserIncludeAttrib parser || attr' == NormalTok
                                                          then ""
-                                                         else " >>= ((withAttribute " ++ show (parserAttribute parser) ++ ") . snd)" 
+                                                         else " >>= ((withAttribute " ++ show attr' ++ ") . snd)"
                                       xs           -> "parseRules " ++ show xs
             "DetectSpaces"     -> "pDetectSpaces"
             "DetectIdentifier" -> "pDetectIdentifier"
@@ -309,16 +333,12 @@ mkSyntaxParser syntax context parser =
                      else empty) <>
                   if parserType parser == "IncludeRules"
                      then mainParser <> char ')'
-                     else (if parserLookAhead parser
-                             then text "lookAhead (" <> mainParser <> text ") >> return ([],\"\") " 
-                             else mainParser <> text " >>= withAttribute " <> 
-                                  text (if null (parserAttribute parser)
-                                           then show (contAttribute context)
-                                           else show (parserAttribute parser))) <> 
-                          char ')' <>
-                          (if parserContext parser `elem` ["", "#stay"]
-                              then empty 
-                              else text " >>~ " <> switchContext (parserContext parser))
+                     else if parserLookAhead parser
+                            then text "lookAhead (" <> mainParser <> char ')'
+                                  <> switchContext (parserContext parser) (text " >> " <>)
+                                  <> text " >> currentContext >>= parseRules)"
+                            else mainParser <> text " >>= withAttribute " <> text (show attr') <> char ')'
+                                  <> switchContext (parserContext parser) (text " >>~ " <>)
       childParsers = parserChildren parser
   in  char '(' <>
       (if null childParsers
@@ -326,13 +346,17 @@ mkSyntaxParser syntax context parser =
           else text "withChildren " <> parserDoc <> char ' ' <> (mkAlternatives $ map (mkSyntaxParser syntax context) childParsers)) <>
       char ')'
 
-switchContext next =
+switchContext :: String -> (Doc -> Doc) -> Doc
+switchContext next finalizer =
   case next of
-     x | "#pop" `isPrefixOf` x -> char '(' <>
-          text (concat $ intersperse " >> " $ replicate (length (filter (=='#') x)) "popContext") <> char ')'
-     "#stay" -> text "return ()"
-     x -> text ("pushContext " ++ show x) 
+     x | "#pop" `isPrefixOf` x -> finalizer $
+          char '(' <> text (concat $ intersperse " >> "
+                                   $ replicate (length (filter (=='#') x)) "popContext") <> char ')'
+     "#stay" -> empty
+     ""      -> empty
+     x       -> finalizer $ text ("pushContext " ++ show x)
 
+langNameToModule :: String -> String
 langNameToModule str =  "Text.Highlighting.Kate.Syntax." ++
   case str of
     "Alerts" -> "Alert"
@@ -355,7 +379,7 @@ compiledRegexName n = "regex_" ++ normalize n
 
 normalize :: String -> String
 normalize "" = ""
-normalize (x:xs) | isAlphaNum x = x : normalize xs 
+normalize (x:xs) | isAlphaNum x = x : normalize xs
 normalize (' ':xs)              = '_':normalize xs
 normalize (x:xs)                = printf "'%2x" (ord x) ++ normalize xs
 
@@ -364,8 +388,8 @@ capitalize (x:xs) = toUpper x : xs
 capitalize [] = []
 
 nameFromPath :: FilePath -> String
-nameFromPath = concat . map capitalize . words . 
-               (map (\c -> if c == '-' then ' ' else c)) . takeFileName . 
+nameFromPath = concat . map capitalize . words .
+               (map (\c -> if c == '-' then ' ' else c)) . takeFileName .
                dropExtension
 
 application :: String -> IOSArrow b SyntaxDefinition
@@ -388,14 +412,14 @@ extractSyntaxDefinition =  proc x -> do
                              lists <- getLists -< x
                              contexts <- getContexts -< x
                              keywordAttr <- getKeywordAttrs -< x
-                             returnA -< SyntaxDefinition { synLanguage      = lang 
+                             returnA -< SyntaxDefinition { synLanguage      = lang
                                                          , synAuthor        = author
                                                          , synVersion       = version
-                                                         , synLicense       = license 
+                                                         , synLicense       = license
                                                          , synExtensions    = sources
                                                          , synCaseSensitive = vBool True caseSensitive
                                                          , synLists         = lists
-                                                         , synContexts      = contexts 
+                                                         , synContexts      = contexts
                                                          , synItemDatas     = itemdatas
                                                          , synKeywordAttr   = if null keywordAttr
                                                                                  then defaultKeywordAttr
@@ -412,12 +436,12 @@ getItemDatas = multi (hasName "itemDatas")
 
 getLists :: IOSArrow XmlTree [(String, [String])]
 getLists = listA $ multi (hasName "list")
-                   >>> 
+                   >>>
                    getAttrValue "name" &&& getListContents
 
 getListContents :: IOSArrow XmlTree [String]
 getListContents = listA $ getChildren
-                          >>> 
+                          >>>
                           hasName "item"
                           >>>
                           getChildren
@@ -438,7 +462,7 @@ getContexts = listA $   multi (hasName "context")
                           fallthroughContext <- getAttrValue "fallthroughContext" -< x
                           dynamic <- getAttrValue "dynamic" -< x
                           parsers <- getParsers -< x
-                          returnA -< SyntaxContext 
+                          returnA -< SyntaxContext
                                            { contName = name
                                            , contAttribute = attribute
                                            , contLineEndContext = if null lineEndContext then "#stay" else lineEndContext
@@ -449,7 +473,7 @@ getContexts = listA $   multi (hasName "context")
                                            , contParsers = parsers }
 
 getParsers :: IOSArrow XmlTree [SyntaxParser]
-getParsers = listA $ getChildren 
+getParsers = listA $ getChildren
                      >>>
                      proc x -> do
                        name <- getName -< x
@@ -465,7 +489,7 @@ getParsers = listA $ getChildren
                        dynamic <- getAttrValue "dynamic" -< x
                        children <- getParsers -< x
                        let tildeRegex = name == "RegExpr" && length str > 0 && head str == '^'
-                       returnA -< SyntaxParser 
+                       returnA -< SyntaxParser
                                     { parserType = name
                                     , parserAttribute = attribute
                                     , parserContext = context
@@ -474,8 +498,8 @@ getParsers = listA $ getChildren
                                     , parserFirstNonSpace = vBool False firstNonSpace
                                     , parserColumn = if tildeRegex
                                                         then Just 0
-                                                        else if null column 
-                                                                then Nothing 
+                                                        else if null column
+                                                                then Nothing
                                                                 else Just (read column)
                                     , parserDynamic = vBool False dynamic
                                     , parserString = if tildeRegex then drop 1 str else str
@@ -490,18 +514,21 @@ getKeywordAttrs = listA $ multi $ hasName "keywords"
                                     caseSensitive <- getAttrValue "casesensitive" -< x
                                     weakDelim <- getAttrValue "weakDeliminator" -< x
                                     additionalDelim <- getAttrValue "additionalDeliminator" -< x
-                                    returnA -< SyntaxKeywordAttr 
+                                    returnA -< SyntaxKeywordAttr
                                                       { keywordCaseSensitive = vBool True caseSensitive
                                                       , keywordDelims = (standardDelims ++ additionalDelim) \\ weakDelim }
 
-standardDelims = " \n\t.():!+,-<=>%&*/;?[]^{|}~\\" 
+standardDelims :: [Char]
+standardDelims = " \n\t.():!+,-<=>%&*/;?[]^{|}~\\"
 
+defaultKeywordAttr :: SyntaxKeywordAttr
 defaultKeywordAttr = SyntaxKeywordAttr { keywordCaseSensitive = True
                                        , keywordDelims = standardDelims }
 
-stripWhitespaceLeft = dropWhile isWhitespace 
-isWhitespace x = elem x [' ', '\t', '\n']
+stripWhitespace :: String -> String
 stripWhitespace = reverse . stripWhitespaceLeft . reverse . stripWhitespaceLeft
+  where stripWhitespaceLeft = dropWhile isWhitespace
+        isWhitespace x = x `elem` [' ', '\t', '\n']
 
 vBool :: Bool -> String -> Bool
 vBool defaultVal value = case value of
@@ -514,7 +541,7 @@ vBool defaultVal value = case value of
 fillTemplate :: Int -> [(String,String)] -> String -> String
 fillTemplate _ _ [] = []
 fillTemplate _ [] lst = lst
-fillTemplate n subs ('\n':xs) = '\n' : fillTemplate 0 subs xs
+fillTemplate _ subs ('\n':xs) = '\n' : fillTemplate 0 subs xs
 fillTemplate n subs ('@':xs) =
   let (pref, suff) = break (=='@') xs
   in  if length pref > 0 && all isAlphaNum pref && length suff > 0
